import ast
import functools
import inspect
import os
import re

import numpy as np

import py_misc_utils.alog as alog
import py_misc_utils.core_utils as pycu
import py_misc_utils.fp_utils as pyf
import py_misc_utils.inspect_utils as pyiu
import py_misc_utils.utils as pyu

from .common_defs import *
from .entity import *
from .emitter import *
from .types import *
from .utils import *
from .vars import *
from .wrap import *


_LOGIC_REMAP = dict(zip('01XUZWHL', '01xxzx10'))

_OPSYMS = {
  ast.Add: OpSym('+'),
  ast.Sub: OpSym('-'),
  ast.Mult: OpSym('*'),
  ast.Div: OpSym('/'),
  ast.Mod: OpSym('%'),
  ast.BitOr: OpSym('|'),
  ast.BitXor: OpSym('^'),
  ast.BitAnd: OpSym('&'),
  ast.MatMult: OpSym('&'),
  ast.Eq: OpSym('=='),
  ast.NotEq: OpSym('!='),
  ast.Lt: OpSym('<'),
  ast.LtE: OpSym('<='),
  ast.Gt: OpSym('>'),
  ast.GtE: OpSym('>='),
  ast.LShift: OpSym('<<'),
  ast.RShift: OpSym('>>'),
}

_FLOAT_OPFNS = {
  ast.Add: 'add',
  ast.Sub: 'sub',
  ast.Mult: 'mul',
  ast.Div: 'div',
  ast.Mod: 'mod',
}


class Verilog_Emitter(Emitter):

  def __init__(self, cfg_file=None, **kwargs):
    super().__init__(cfg_file=cfg_file, **kwargs)
    self.kind = 'verilog'
    self.file_ext = '.sv'
    self.eol = ';'

    vfpu_path = pyu.dict_rget(self._cfg, 'backend/verilog/vfpu_conf',
                              defval='hdl_libs/verilog/vfpu.yaml')

    self._vfpu = self.load_extern_module(vfpu_path)
    self._init_module_places()

  @staticmethod
  def fpmod_resolve(modname, fnname, argref, **kwargs):
    def resolver(ctx, cargs, rkwargs):
      xkwargs = kwargs.copy()
      xkwargs.update(rkwargs)
      return ctx.emitter._fpmod_resolve(modname, fnname, cargs, argref, xkwargs)

    return resolver

  def _fpmod_resolve(self, modname, fnname, cargs, argref, kwargs):
    xmod = self._get_extern_module(modname)

    dtype = self._argref_dtype(argref, cargs, kwargs)

    fspec = self.float_spec(dtype)
    kwargs.update(FPEXP=fspec.exp, FPMANT=fspec.mant)

    return self._call_external_module(xmod, fnname, dtype, *cargs, **kwargs)

  def _init_module_places(self):
    self.module_vars_place = self.emit_placement()
    self._modules_place = self.emit_placement()
    self._entity_place = self.emit_placement()

  def _mod_call(self, fnname, dtype, *args, **kwargs):
    return self._call_external_module(self._vfpu, fnname, dtype, *args, **kwargs)

  def _scalar_remap(self, value):
    if isinstance(value, bool):
      return '1' if value else '0'

  def _type_of(self, dtype):
    kind = 'logic'
    nbits, shape = dtype.nbits, dtype.array_shape

    adims = ''.join(f'[{x}]' for x in shape)
    if isinstance(dtype, Uint):
      if shape:
        return f'{kind} [{nbits - 1}: 0] {{}}{adims}'
      else:
        return f'{kind} [{nbits - 1}: 0] {{}}'
    elif isinstance(dtype, Sint):
      if shape:
        return f'{kind} signed [{nbits - 1}: 0] {{}}{adims}'
      else:
        return f'{kind} signed [{nbits - 1}: 0] {{}}'
    elif isinstance(dtype, Bits):
      if nbits == 1:
        return f'{kind} {{}}{adims}' if shape else f'{kind} {{}}'

      if shape:
        return f'{kind} [{nbits - 1}: 0] {{}}{adims}'
      else:
        return f'{kind} [{nbits - 1}: 0] {{}}'
    elif isinstance(dtype, Bool):
      return f'{kind} {{}}{adims}' if shape else f'{kind} {{}}'
    elif isinstance(dtype, Integer):
      return f'integer {{}}{adims}' if shape else f'integer {{}}'
    elif isinstance(dtype, Real):
      return f'real {{}}{adims}' if shape else f'real {{}}'
    elif isinstance(dtype, Float):
      fspec = self.float_spec(dtype)
      nfbits = 1 + fspec.exp + fspec.mant
      if shape:
        return f'{kind} [{nfbits - 1}: 0] {{}}{adims}'
      else:
        return f'{kind} [{nfbits - 1}: 0] {{}}'

    fatal(f'Unknown type: {dtype}', exc=TypeError)

  def _resize_bits(self, value, nbits):
    if value.dtype.nbits == nbits:
      return value

    result = f'{nbits}\'({self.svalue(value)})'

    shape = list(value.dtype.shape)
    shape[-1] = nbits

    return value.new_value(result, shape=shape)

  def _to_bool(self, value, dtype):
    if isinstance(value, Value):
      xvalue = self.svalue(value)
      if isinstance(value.dtype, (Sint, Uint, Bits)):
        return f'&{paren(xvalue)}'

      zero = self._cast(0, value.dtype)
      return f'{paren(xvalue)} != {zero}'

    if isinstance(value, str):
      value = pycu.infer_value(value)

    return '1' if value else '0'

  def _to_int(self, value, dtype, signed):
    itype = 'signed' if signed else 'unsigned'
    if isinstance(value, Value):
      if isinstance(value.dtype, (Uint, Sint)):
        xvalue = self.svalue(self._resize_bits(value, dtype.nbits))
        return f'{itype}\'({xvalue})' if type(dtype) != type(value.dtype) else xvalue
      elif isinstance(value.dtype, Bits):
        xvalue = self.svalue(self._resize_bits(value, dtype.nbits))
        return f'signed\'({xvalue})' if signed else xvalue
      elif isinstance(value.dtype, Float):
        fspec = self.float_spec(dtype)

        mcall = self._mod_call('to_integer', dtype, value,
                               FPEXP=fspec.exp,
                               FPMANT=fspec.mant,
                               INTSIZE=dtype.nbits)

        xvalue = f'{dtype.nbits}\'({mcall})'
        return f'signed\'({xvalue})' if signed else xvalue
      elif isinstance(value.dtype, Integer):
        xvalue = f'{dtype.nbits}\'({self.svalue(value)})'
        return f'signed\'({xvalue})' if signed else xvalue
      elif isinstance(value.dtype, Real):
        xvalue = f'{dtype.nbits}\'($rtoi({self.svalue(value)}))'
        return f'signed\'({xvalue})' if signed else xvalue
      elif isinstance(value.dtype, Bool):
        xvalue = f'{dtype.nbits}\'({self.svalue(value)})'
        return f'signed\'({xvalue})' if signed else xvalue
      else:
        fatal(f'Unknown type: {value.dtype}', exc=TypeError)

    if isinstance(value, str):
      value = pycu.infer_value(value)

    xvalue = f'{dtype.nbits}\'({int(value)})'
    return f'signed\'({xvalue})' if signed else xvalue

  def _to_uint(self, value, dtype):
    return self._to_int(value, dtype, False)

  def _to_sint(self, value, dtype):
    return self._to_int(value, dtype, True)

  def _try_convert_literal(self, value):
    # This API should not try to convert Python `int` or `float` literals, as those
    # are better off directly going to the cast operators, resulting in less bits
    # resize operations.
    if value is True or value is False:
      return Value(BOOL, '1' if value else '0')
    if value is None:
      return Value(VOID)
    if isinstance(value, str):
      bvalue = bitstring(value, remap=lambda x: _LOGIC_REMAP[x])
      if bvalue is not None:
        return bvalue.new_value(f'{bvalue.dtype.nbits}\'b{bvalue.value}')

      dtype, ivalue = self._match_intstring(value)
      if dtype is not None:
        if isinstance(dtype, Uint):
          return Value(dtype, f'{dtype.nbits}\'({ivalue})')
        else:
          return Value(dtype, f'signed\'({dtype.nbits}\'({ivalue}))')

    return value

  def _to_bits(self, value, dtype):
    if isinstance(value, Value):
      if isinstance(value.dtype, (Uint, Sint)):
        xvalue = self.svalue(self._resize_bits(value, dtype.nbits))
        return f'unsigned\'({xvalue})' if isinstance(value.dtype, Sint) else xvalue
      elif isinstance(value.dtype, Bits):
        return self.svalue(self._resize_bits(value, dtype.nbits))
      elif isinstance(value.dtype, Float):
        fspec = self.float_spec(dtype)

        mcall = self._mod_call('to_integer', dtype, value,
                               FPEXP=fspec.exp,
                               FPMANT=fspec.mant,
                               INTSIZE=dtype.nbits)

        return f'{dtype.nbits}\'({mcall})'
      elif isinstance(value.dtype, Integer):
        return f'{dtype.nbits}\'({self.svalue(value)})'
      elif isinstance(value.dtype, Real):
        return f'{dtype.nbits}\'($rtoi({self.svalue(value)}))'
      elif isinstance(value.dtype, Bool):
        return f'{dtype.nbits}\'({self.svalue(value)})'
      else:
        fatal(f'Unable to convert to {dtype}: {value.dtype}')

    if isinstance(value, str):
      if value.startswith('0b'):
        value = value[2: ]
      if dtype.nbits > len(value):
        bvalue = ('0' * (dtype.nbits - len(value))) + value
      else:
        bvalue = value[-dtype.nbits: ]

      bstr = ''.join(_LOGIC_REMAP[x] for x in bvalue.upper())

      return f'{len(bstr)}\'b{bstr}'
    elif isinstance(value, int) and abs(value) >= 2**31:
      return f'{dtype.nbits}\'b{self._int2bits(value, dtype.nbits)}'

    return f'{dtype.nbits}\'({value})'

  def _to_float(self, value, dtype):
    fspec = self.float_spec(dtype)
    if isinstance(value, Value):
      if isinstance(value.dtype, (Uint, Sint, Bits)):
        return self._mod_call('from_integer', dtype, value,
                              FPEXP=fspec.exp,
                              FPMANT=fspec.mant,
                              INTSIZE=value.dtype.nbits)
      elif isinstance(value.dtype, Float):
        ifspec = self.float_spec(value.dtype)

        return self._mod_call('convert', dtype, value,
                              FPEXP=ifspec.exp,
                              FPMANT=ifspec.mant,
                              OUT_FPEXP=fspec.exp,
                              OUT_FPMANT=fspec.mant)
      elif isinstance(value.dtype, Integer):
        nbits = max(32, dtype.nbits)

        return self._mod_call('from_integer', dtype, f'{nbits}\'({self.svalue(value)})',
                              FPEXP=fspec.exp,
                              FPMANT=fspec.mant,
                              INTSIZE=nbits)
      elif isinstance(value.dtype, Real):
        return self._mod_call('from_real', dtype, value,
                              FPEXP=fspec.exp,
                              FPMANT=fspec.mant)
      elif isinstance(value.dtype, Bool):
        mcall_one = self._mod_call('one', dtype,
                                   FPEXP=fspec.exp,
                                   FPMANT=fspec.mant)
        mcall_zero = self._mod_call('zero', dtype,
                                    FPEXP=fspec.exp,
                                    FPMANT=fspec.mant)

        xvalue = self.svalue(value)
        return f'{paren(xvalue)} ? {mcall_one}() : {mcall_zero}()'

    if isinstance(value, str):
      value = pycu.infer_value(value)

    fbits = pyf.real_to_packedbits(float(value), fspec.exp, fspec.mant)

    return f'{dtype.nbits}\'b{fbits:0{dtype.nbits}b}'

  def _to_integer(self, value, dtype):
    if isinstance(value, Value):
      if value.dtype == dtype:
        return self.svalue(value)
      elif isinstance(value.dtype, (Bool, Sint, Uint, Bits)):
        return f'int\'({self.svalue(value)})'
      elif isinstance(value.dtype, Float):
        fspec = self.float_spec(value.dtype)

        mcall = self._mod_call('to_integer', dtype, value,
                               FPEXP=fspec.exp,
                               FPMANT=fspec.mant,
                               INTSIZE=32)

        return f'int\'({mcall})'
      elif isinstance(value.dtype, Real):
        return f'int\'({self.svalue(value)})'
      else:
        fatal(f'Unable to convert to integer: {value} {dtype}')

    return str(value) if isinstance(value, int) else f'int\'({value})'

  def _to_real(self, value, dtype):
    if isinstance(value, Value):
      if value.dtype == dtype:
        return self.svalue(value)
      elif isinstance(value.dtype, (Bool, Sint, Uint, Bits, Integer)):
        return f'real\'({self.svalue(value)})'
      elif isinstance(value.dtype, Float):
        fspec = self.float_spec(value.dtype)

        return self._mod_call('to_real', dtype, value,
                              FPEXP=fspec.exp,
                              FPMANT=fspec.mant)
      else:
        fatal(f'Unable to convert to real: {value} {dtype}')

    return str(value) if isinstance(value, float) else f'real\'({value})'

  def _scalar_cast(self, value, dtype):
    if isinstance(dtype, Bool):
      return self._to_bool(value, dtype)
    if isinstance(dtype, Uint):
      return self._to_uint(value, dtype)
    if isinstance(dtype, Sint):
      return self._to_sint(value, dtype)
    if isinstance(dtype, Bits):
      return self._to_bits(value, dtype)
    if isinstance(dtype, Float):
      return self._to_float(value, dtype)
    if isinstance(dtype, Integer):
      return self._to_integer(value, dtype)
    if isinstance(dtype, Real):
      return self._to_real(value, dtype)

    fatal(f'Unknown type: {dtype}')

  def _do_cast(self, value, dtype):
    if not isinstance(value, Value):
      value = self._try_convert_literal(value)

    if isinstance(value, np.ndarray):
      shape = dtype.array_shape
      if shape != tuple(value.shape):
        fatal(f'Shape mismatch: {tuple(value.shape)} vs. {shape}')

      element_type = dtype.element_type()
      parts = []
      for idx in np.ndindex(shape):
        parts.append(self._scalar_cast(value[idx].item(), element_type))

      xvalue = flat2shape(parts, shape, '{', '}')
    elif isinstance(value, Value) and value.dtype.array_shape:
      shape, vshape = dtype.array_shape, value.dtype.array_shape
      if shape != vshape:
        fatal(f'Shape mismatch: {vshape} vs. {shape}')

      element_type = dtype.element_type()
      velement_type = value.dtype.element_type()
      avalue = paren(self.svalue(value))
      parts = []
      for idx in np.ndindex(shape):
        substr = ''.join(f'[{x}]' for x in idx)
        svalue = Value(velement_type, f'{avalue}{substr}')
        parts.append(self._scalar_cast(svalue, element_type))

      xvalue = flat2shape(parts, shape, '{', '}')
    else:
      xvalue = self._scalar_cast(unwrap(value), dtype)
      for size in reversed(dtype.array_shape):
        xvalue = f'\'{{{size}{{{xvalue}}}}}'

    return xvalue

  def _cast(self, value, dtype):
    if isinstance(value, Value) and value.dtype == dtype:
      return self.svalue(value)

    return self._do_cast(value, dtype)

  def cast(self, value, dtype, isreg=None):
    if isinstance(value, Value) and value.dtype == dtype:
      return value

    xvalue = self._do_cast(value, dtype)
    if isreg is None and isinstance(value, Value):
      isreg = value.isreg

    return Value(dtype, xvalue, isreg=isreg)

  def eval_tostring(self, value):
    xvalue = self.svalue(value)
    if isinstance(value, Value):
      if isinstance(value.dtype, (Uint, Sint, Integer)):
        return f'$sformatf("%d", {xvalue})'
      elif isinstance(value.dtype, (Bool, Bits)):
        return f'$sformatf("%b", {xvalue})'
      elif isinstance(value.dtype, Real):
        return f'$sformatf("%e", {xvalue})'
      elif isinstance(value.dtype, Float):
        fspec = self.float_spec(value.dtype)

        mcall = self._mod_call('to_real', Real(), value,
                               FPEXP=fspec.exp,
                               FPMANT=fspec.mant)

        return f'$sformatf("%e", {mcall})'
      else:
        fatal(f'Unable to convert to string: {value}')

    return self.quote_string(xvalue)

  def quote_string(self, s):
    es = s.replace('"', '\\"')
    return f'"{es}"'

  def eval_token(self, token):
    if token == 'NOW':
      time_fmt = os.getenv('TIME_FMT', '%0t')
      return f'$sformatf("{time_fmt}", $time)'

  def emit_finish(self):
    self.emit_code('$finish;')

  def emit_wait_for(self, ts=None):
    if ts is not None:
      wts, tu = self._normalize_time(ts)
      self.emit_code(f'#{wts}{tu};')
    else:
      self.emit_code('forever;')

  def emit_wait_rising(self, arg):
    self.emit_code(f'@(posedge {paren(self.svalue(arg))});')

  def emit_wait_falling(self, arg):
    self.emit_code(f'@(negedge {paren(self.svalue(arg))});')

  def emit_wait_until(self, arg):
    self.emit_code(f'@({paren(self.svalue(arg))});')

  def emit_report(self, parts, severity=None):
    self.emit_write(parts)

  def emit_write(self, parts):
    self._emit_line('$display("' + ('%s' * len(parts)) + '", ' + ', '.join(parts) + ');')

  def _gen_array_access(self, arg, idx):
    idx = pyu.as_sequence(idx)

    ashape = arg.dtype.shape
    if len(idx) > len(ashape):
      fatal(f'Wrong indexing for shape: {idx} vs. {ashape}')

    shape, coords = [], []
    for i, ix in enumerate(idx):
      if isinstance(ix, slice):
        if isinstance(ix.start, Value):
          if ix.stop is not None:
            fatal(f'Variable part select ({arg} [{i}]) slice stop must be empty: {ix.stop}')
          if ix.step is None:
            fatal(f'Variable part select ({arg} [{i}]) slice step must not be empty')

          step = abs(ix.step)
          if step > ashape[i]:
            fatal(f'Variable part select ({arg} [{i}]) is too big: {step} ({ashape[i]})')

          base = self._to_integer(ix.start, Integer())

          if (i == len(ashape) - 1) == (ix.step >= 0):
            coords.append(f'{paren(base)} + {step - 1} -: {step}')
          else:
            coords.append(f'{paren(base)} +: {step}')

          shape.append(step)
        else:
          step = ix.step if ix.step is not None else 1
          if abs(step) != 1:
            fatal(f'Slice step must be 1: {step}')

          start, stop = validate_slice(ix.start, ix.stop, ashape[i], step)

          if i == len(ashape) - 1:
            coords.append(f'{stop - step}: {start}')
          else:
            coords.append(f'{start}: {stop - step}')

          shape.append(abs(stop - start))
      else:
        if isinstance(ix, Value):
          if not isinstance(ix.dtype, Integer):
            ix = self._to_integer(ix, Integer())
        elif isinstance(ix, int) and ix < 0:
          ix = ashape[i] + ix

        coords.append(self.svalue(ix))
        shape.append(1)

    shape = pyu.squeeze(shape + list(arg.dtype.full_shape[len(idx): ]), keep_dims=1,
                        sdir=pyu.MAJOR)
    avalue = paren(self.svalue(arg), kind='{}') + ''.join(f'[{x}]' for x in coords)

    return avalue, shape

  def flush(self):
    return self._load_libs() + self._expand()

  def is_root_variable(self, var):
    return var.isreg or var.is_const()

  def var_remap(self, var, is_store):
    return var

  def _emit_attributes(self, attributes):
    attrs = []
    for aname, avalue in self._enum_attributes(attributes):
      if isinstance(avalue, str):
        attrs.append(f'{aname} = "{avalue}"')
      else:
        attrs.append(f'{aname} = {avalue}')

    if attrs:
      self._emit_line('(* ' + ', '.join(attrs) + ' *)')

  def emit_declare_variable(self, name, var):
    if var.is_const():
      vprefix, is_const = 'const ', True
    else:
      vprefix, is_const = '', False

    vinit = ''
    if var.init is not None:
      vinit = f' = {self._cast(var.init, var.dtype)}'

    # Wires with process scope are always automatic (never should latch).
    if not self.is_root_variable(var) and self._proc.kind != ROOT_PROCESS:
      vprefix = f'automatic {vprefix}'

    ntype = self._type_of(var.dtype).format(name)

    vspec = var.vspec
    if vspec is not None and vspec.attributes is not None:
      self._emit_attributes(vspec.attributes)

    self._emit_line(f'{vprefix}{ntype}{vinit};')

  def emit_assign(self, var, name, value):
    xvalue = self._cast(value, var.dtype)

    delay = self.get_context('delay')
    xdelay = f'#{paren(self.svalue(delay))} ' if delay is not None else ''

    cont_assign = 'assign ' if self._proc.kind == ROOT_PROCESS else ''
    # Sequential designs (processes having posedge/negedge sensitivity) should use non
    # blocking assignments.
    asop = '=' if cont_assign or not self._edge_inputs or not var.isreg else '<='

    self._emit_line(f'{xdelay}{cont_assign}{var.value} {asop} {xvalue};')

  def make_port_arg(self, port_arg):
    return port_arg

  def emit_entity(self, ent, kwargs, ent_name=None):
    if ent_name is None:
      ent_name = pyiu.cname(ent)

    iname = self._get_entity_inst(ent_name)

    eparams, params = kwargs.pop(PARAM_KEY, None), ''
    if eparams:
      params = '#(' + ', '.join(f'.{k}({v})' for k, v in eparams.items()) + ') '

    with self.placement(self._entity_place):
      self._emit_line(f'{ent_name} {params}{iname}(')
      with self.indent():
        binds = []
        for i, pin in enumerate(ent.PORTS):
          arg = kwargs[pin.name]

          if pin.is_ifc():
            xargs = pin.ifc_expand(arg)
          else:
            xargs = ((pin, arg),)

          for xpin, xarg in xargs:
            if not isinstance(xarg, Value):
              fatal(f'Argument must be a Value subclass: {xarg}')

            if xarg.is_none():
              binds.append(f'.{xpin.name}()')
            else:
              earg = self.svalue(xarg)
              binds.append(f'.{xpin.name}({earg})')

        self._emit_lines(binds, sep=',')

      self._emit_line(f');')

  def emit_module_def(self, name, ent, comment=None):
    self._mod_comment = comment

  def emit_module_decl(self, name, ent):
    if self._mod_comment:
      self.emit_comment(self._mod_comment)

    xports = ent.expanded_ports()

    self._emit_line(f'module {name}(' + ', '.join(ap.port.name for ap in xports) + ');')
    with self.indent():
      for ap in xports:
        pin, arg = ap.port, ap.arg

        pdir = 'input' if pin.is_ro() else 'output' if pin.is_wo() else 'inout'
        ntype = self._type_of(arg.dtype).format(pin.name)

        self._emit_line(f'{pdir} {ntype};')

      self._init_module_places()

  def emit_module_end(self):
    with self.placement(self._modules_place):
      for iid, inst in self._itor:
        params = [f'.{k}({v})' for k, v in inst.params.items()]
        args = [f'.{k}({self.svalue(v)})' for k, v in inst.args.items()]

        self._emit_line(f'{inst.name} #(' + ', '.join(params) + f') {iid}(' + ', '.join(args) + ');')

    self._emit_line(f'endmodule')
    self._module_reset()

  def emit_process_decl(self, name, sensitivity=None, process_kind=None,
                        process_args=None):
    self._process_init(name, process_kind, process_args, sensitivity)

    if process_kind == INIT_PROCESS:
      if sensitivity:
        fatal(f'Sensitivity list not allowed in init process')

      self._emit_line('initial')
    elif sensitivity:
      conds = []
      for sname, sens in sensitivity.items():
        if sens.trigger == POSEDGE:
          conds.append(f'posedge {paren(sname)}')
          self._edge_inputs.append((sname, sens))
        elif sens.trigger == NEGEDGE:
          conds.append(f'negedge {paren(sname)}')
          self._edge_inputs.append((sname, sens))
        else:
          conds.append(paren(sname))

      if len(conds) > len(self._edge_inputs):
        self._emit_line(f'always @(' + ' or '.join(conds) + ')')
      else:
        self._emit_line(f'always_ff @(' + ' or '.join(conds) + ')')
    else:
      proc_mode = process_args.get('proc_mode') if process_args else None
      if proc_mode == 'comb':
        self._emit_line('always_comb')
      elif proc_mode == 'latch':
        self._emit_line('always_latch')
      elif proc_mode == 'loop':
        self._emit_line('always')
      else:
        self._emit_line('always @(*)')

  def emit_process_begin(self):
    if self._proc.name:
      self._emit_line(f'{self._proc.name} : begin')
    else:
      self._emit_line(f'begin')
    self.process_vars_place = self.emit_placement(extra_indent=1)

  def _process_reset(self):
    super()._process_reset()
    self._edge_inputs = []

  def emit_process_end(self):
    self._emit_line(f'end')
    self._process_reset()

  def emit_comment(self, msg):
    for ln in msg.split('\n'):
      self._emit_line(f'// {ln}')

  def emit_If(self, test):
    xtest = self._cast(test, BOOL)
    self._emit_line(f'if {paren(xtest)} begin')

  def emit_Elif(self, test):
    xtest = self._cast(test, BOOL)
    self._emit_line(f'end else if {paren(xtest)} begin')

  def emit_Else(self):
    self._emit_line(f'end else begin')

  def emit_EndIf(self):
    self._emit_line(f'end')

  def emit_For(self, vname, start, end, step):
    if step > 0:
      loop_cmp, loop_incr = '<=', '+='
    else:
      loop_cmp, loop_incr = '>=', '-='

    vtype = 'genvar' if self._proc.kind == ROOT_PROCESS else 'int'

    self._emit_line(f'for ({vtype} {vname} = {start}; {vname} {loop_cmp} {end}; ' \
                    f'{vname} {loop_incr} {abs(step)}) begin')

  def emit_EndFor(self):
    self._emit_line(f'end')

  def emit_Break(self):
    self._emit_line(f'break;')

  def emit_Continue(self):
    self._emit_line(f'continue;')

  def emit_Assert(self, test, parts):
    xtest = self.svalue(test)
    if parts:
      self._emit_line(f'assert {xtest} else $error("' + ('%s' * len(parts)) + '", ' + ', '.join(parts) + '");')
    else:
      self._emit_line(f'assert {xtest};')

  def emit_match_cases(self, subject, cases):
    xsubject = self.svalue(subject)
    self._emit_line(f'case ({xsubject})')

    with self.indent():
      for mc in cases:
        if mc.pattern is not None:
          xpattern = self._cast(mc.pattern, subject.dtype)
          self._emit_line(f'{paren(xpattern)}: begin')
        else:
          self._emit_line(f'default: begin')

        with self.indent():
          self._emit(mc.scope)
        self._emit_line(f'end')

    self._emit_line(f'endcase')

  def _build_op(self, op, left, right):
    sop = _OPSYMS[pyiu.classof(op)]
    if sop.isfn:
      return f'{sop.sym}({paren(left)}, {paren(right)})'
    else:
      return f'{paren(left)} {sop.sym} {paren(right)}'

  def _build_arith_op(self, op, left, right, dtype):
    if isinstance(dtype, Float):
      fspec = self.float_spec(dtype)
      opfn = _FLOAT_OPFNS[pyiu.classof(op)]

      return self._mod_call(opfn, dtype, left, right,
                            FPEXP=fspec.exp,
                            FPMANT=fspec.mant)
    else:
      return self._build_op(op, left, right)

  def _arith_ctype_cast(self, arg, dtype):
    if isinstance(arg, int):
      if abs(arg) >= 2**31 or type(dtype) not in {Uint, Sint, Real, Integer}:
        return Value(dtype, self._cast(arg, dtype))
    elif isinstance(arg, float):
      if type(dtype) not in {Real}:
        return Value(dtype, self._cast(arg, dtype))

    return Value(dtype, dtype.ctype(arg))

  def eval_BinOp(self, op, left, right):
    if isinstance(op, (ast.Add, ast.Sub, ast.Mult, ast.Div, ast.Mod)):
      left, right = self._marshal_arith_op([left, right],
                                           ctype_cast=self._arith_ctype_cast)
      xleft, xright = self.svalue(left), self.svalue(right)

      alog.debug(lambda: f'\tBinOp: {xleft}\t{pyiu.cname(op)}\t{xright}')
      result = self._build_arith_op(op, xleft, xright, left.dtype)
      # The signed/unsigned multiplication result has a number of bits which is the
      # sum of the ones of the operands, which is not the behaviour we want.
      if isinstance(op, ast.Mult) and isinstance(left.dtype, (Sint, Uint)):
        result = f'{left.dtype.nbits}\'({result})'

      return Value(left.dtype, result)
    elif isinstance(op, (ast.LShift, ast.RShift)):
      left, right = self._marshal_shift_op(left, right)
      xleft, xright = self.svalue(left), self.svalue(right)

      alog.debug(lambda: f'\tBinOp: {xleft}\t{pyiu.cname(op)}\t{xright}')

      return Value(left.dtype, self._build_op(op, xleft, xright))
    elif isinstance(op, (ast.BitOr, ast.BitXor, ast.BitAnd)):
      left, right = self._marshal_bit_op([left, right])
      xleft, xright = self.svalue(left), self.svalue(right)

      alog.debug(lambda: f'\tBinOp: {xleft}\t{pyiu.cname(op)}\t{xright}')

      return Value(left.dtype, self._build_op(op, xleft, xright))
    elif isinstance(op, ast.MatMult):
      # Steal MatMult ('@') for concatenation!
      dtype, (left, right) = self._marshal_concat_op([left, right])
      xleft, xright = self.svalue(left), self.svalue(right)

      alog.debug(lambda: f'\tBinOp: {xleft}\t{pyiu.cname(op)}\t{xright}')

      return Value(dtype, f'{{{xleft}, {xright}}}')
    else:
      fatal(f'Unsupported operation: {op}')

  def eval_UnaryOp(self, op, arg):
    xvalue = self.svalue(arg)

    alog.debug(lambda: f'\tUnaryOp: {pyiu.cname(op)}\t{xvalue}')

    if isinstance(op, ast.UAdd):
      # Unary addition is a noop for HDL data types.
      result = xvalue
    else:
      if isinstance(arg, Value) and isinstance(arg.dtype, Float):
        if isinstance(op, ast.USub):
          fspec = self.float_spec(arg.dtype)

          result = self._mod_call('neg', arg.dtype, arg,
                                  FPEXP=fspec.exp,
                                  FPMANT=fspec.mant)
        else:
          fatal(f'Unsupported operation for type {arg.dtype}: {op}')
      else:
        if isinstance(op, ast.USub):
          result = f'-{paren(xvalue)}'
        elif isinstance(op, ast.Not):
          result = f'!{paren(xvalue)}'
        elif isinstance(op, ast.Invert):
          result = f'~{paren(xvalue)}'
        else:
          fatal(f'Unsupported operation for type {arg.dtype}: {op}')

    return Value(arg.dtype, result)

  def eval_BoolOp(self, op, args):
    xargs = [self._cast(a, BOOL) for a in args]

    alog.debug(lambda: f'\tBoolOp: {pyiu.cname(op)}\t{pyu.stri(xargs)}')

    if isinstance(op, ast.And):
      result = self._paren_join(' && ', xargs)
    elif isinstance(op, ast.Or):
      result = self._paren_join(' || ', xargs)
    else:
      fatal(f'Unsupported operation: {op}')

    return Value(BOOL, result)

  def eval_Compare(self, left, ops, comps):
    comps = self._marshal_compare_op([left] + list(comps))
    xcomps = [self.svalue(comp) for comp in comps]

    alog.debug(lambda: f'\tCompare: {[pyiu.cname(x) for x in ops]}\t{pyu.stri(xcomps)}')

    results = []
    for i, op in enumerate(ops):
      cres = self._build_op(op, xcomps[i], xcomps[i + 1])
      results.append(cres)

    result = self._paren_join(' && ', results)

    return Value(BOOL, result)

  def eval_Subscript(self, arg, idx):
    result, shape = self._gen_array_access(arg, idx)

    return arg.new_value(result, shape=shape, keepref=True)

  def eval_IfExp(self, test, body, orelse):
    xtest = self.svalue(test)
    body, orelse = self._marshal_ifexp_op([body, orelse])
    xbody, xorelse = self.svalue(body), self.svalue(orelse)

    alog.debug(lambda: f'\tIfExp: {xtest} ? {xbody} : {xorelse}')

    return Value(body.dtype, f'{paren(xtest)} ? {xbody} : {xorelse}')

  # Extension functions.
  def eval_is_nan(self, value):
    if not isinstance(value.dtype, Float):
      fatal(f'Unsupported type: {value.dtype}')

    fspec = self.float_spec(value.dtype)

    result = self._mod_call('is_nan', BOOL, value,
                            FPEXP=fspec.exp,
                            FPMANT=fspec.mant)

    return Value(BOOL, result)

  def eval_is_inf(self, value):
    if not isinstance(value.dtype, Float):
      fatal(f'Unsupported type: {value.dtype}')

    fspec = self.float_spec(value.dtype)

    result = self._mod_call('is_inf', BOOL, value,
                            FPEXP=fspec.exp,
                            FPMANT=fspec.mant)

    return Value(BOOL, result)


# Register Verilog (SystemVerilog >= 2012) emitter class.
Emitter.register('verilog', Verilog_Emitter)




import re

import py_misc_utils.core_utils as pycu
import py_misc_utils.utils as pyu


def _add_inputs(inseq, gglobals, ddict):
  for ein in inseq:
    names, expr = re.split(r'\s*=\s*', ein, maxsplit=1)

    try:
      value = eval(expr, gglobals)
    except:
      value = expr

    for name in pyu.comma_split(names):
      pycu.dict_add(ddict, name, value)


def parse_kwargs(kwargs, gglobals, ddict=None):
  ddict = pyu.value_or(ddict, dict())
  if kwargs:
    _add_inputs(kwargs, gglobals, ddict)

  return ddict


def parse_inputs(inputs, kwargs, gglobals):
  ddict = dict()
  if inputs:
    _add_inputs(inputs, gglobals, ddict)

  return parse_kwargs(kwargs, gglobals, ddict=ddict)


def parse_args(cfgfile, args):
  cfg = pyu.load_config(args.cfgfile)
  for k, v in cfg.items():
    setattr(args, k, v)




import argparse
import os
import re
import unittest

import py_misc_utils.alog as alog
import py_misc_utils.app_main as app_main
import py_misc_utils.core_utils as pycu
import py_misc_utils.utils as pyu

import pyxhdl as X

import test_utils as tu


def _main(args):
  test_folder = (args.test_folder or
                 os.environ.get('TESTS_FOLDER') or
                 os.path.dirname(os.path.abspath(__file__)))

  alog.debug(f'TestFolder: {test_folder}')

  for arg in args.backed_arg or ():
    m = re.match(r'([^\s]+)\s*:\s*([^\s]+)(\s*=\s*([^\s]+))?', arg)
    if not m:
      pyu.fatal(f'Invalid backend argument: {arg}')

    value = pycu.infer_value(m.group(4)) if m.group(4) else None
    tu.add_backend_arg(m.group(1).lower(), m.group(2), value)

  loader = unittest.TestLoader()
  tests = loader.discover(test_folder, pattern=args.files)
  runner = unittest.runner.TextTestRunner(verbosity=args.verbosity)
  runner.run(tests)


if __name__ == '__main__':
  parser = argparse.ArgumentParser(description='PyXHDL Tests Runner',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument('--verbosity', type=int, default=2,
                      help='The test verbosity')
  parser.add_argument('--files', type=str, default='t_*.py',
                      help='The pattern to match files whose tests need to be run')
  parser.add_argument('--test_folder',
                      help='The folder where the test files are stored')
  parser.add_argument('--backed_arg', nargs='*',
                      help='An argument for the backend (ie, "vhdl:var=value")')

  app_main.main(parser, _main)




import ast
import collections
import functools
import inspect
import os
import re

import numpy as np

import py_misc_utils.alog as alog
import py_misc_utils.core_utils as pycu
import py_misc_utils.inspect_utils as pyiu
import py_misc_utils.utils as pyu

from .common_defs import *
from .entity import *
from .emitter import *
from .types import *
from .utils import *
from .vars import *
from .wrap import *


_STD_HEADER = """library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;
use ieee.float_pkg.all;
use std.textio.all;

library work;
use work.all;
"""

_OPSYMS = {
  ast.Add: OpSym('+'),
  ast.Sub: OpSym('-'),
  ast.Mult: OpSym('*'),
  ast.Div: OpSym('/'),
  ast.Mod: OpSym('mod'),
  ast.BitOr: OpSym('or'),
  ast.BitXor: OpSym('xor'),
  ast.BitAnd: OpSym('and'),
  ast.MatMult: OpSym('&'),
  ast.Eq: OpSym('='),
  ast.NotEq: OpSym('/='),
  ast.Lt: OpSym('<'),
  ast.LtE: OpSym('<='),
  ast.Gt: OpSym('>'),
  ast.GtE: OpSym('>='),
  ast.LShift: OpSym('pyxhdl.bit_shl', True),
  ast.RShift: OpSym('pyxhdl.bit_shr', True),
}


class VHDL_Emitter(Emitter):

  def __init__(self, cfg_file=None, **kwargs):
    super().__init__(cfg_file=cfg_file, **kwargs)
    self.kind = 'vhdl'
    self.file_ext = '.vhd'
    self.eol = ';'
    self._arch = self._cfg.get('backend/vhdl/entity_arch', 'behavior')
    self._mod_attributes = dict()
    self._proc_indent = 0
    self.module_vars_place = self.emit_placement()
    self._modules_place = self.emit_placement()
    self._entity_place = self.emit_placement()

  def _emit_header(self):
    hdr = self._cfg.get('backend/vhdl/header', _STD_HEADER)
    for ln in hdr.split('\n'):
      self._emit_line(ln)

  def _scalar_remap(self, value):
    if isinstance(value, bool):
      return 'true' if value else 'false'

  def _type_of(self, dtype):
    nbits, shape = dtype.nbits, dtype.array_shape

    adims = ''.join(f'(0 to {x - 1})' for x in shape)
    if isinstance(dtype, Uint):
      if shape:
        return f'pyxhdl.uint_array{len(shape)}d{adims}({nbits - 1} downto 0)'
      else:
        return f'unsigned({nbits - 1} downto 0)'
    elif isinstance(dtype, Sint):
      if shape:
        return f'pyxhdl.sint_array{len(shape)}d{adims}({nbits - 1} downto 0)'
      else:
        return f'signed({nbits - 1} downto 0)'
    elif isinstance(dtype, Bits):
      if nbits == 1:
        return f'pyxhdl.slv_array{len(shape)}d{adims}' if shape else 'std_logic'

      if shape:
        return f'pyxhdl.bits_array{len(shape)}d{adims}({nbits - 1} downto 0)'
      else:
        return f'std_logic_vector({nbits - 1} downto 0)'
    elif isinstance(dtype, Bool):
      return f'pyxhdl.bool_array{len(shape)}d{adims}' if shape else 'boolean'
    elif isinstance(dtype, Integer):
      return f'pyxhdl.integer_array{len(shape)}d{adims}' if shape else 'integer'
    elif isinstance(dtype, Real):
      return f'pyxhdl.real_array{len(shape)}d{adims}' if shape else 'real'
    elif isinstance(dtype, Float):
      fspec = self.float_spec(dtype)
      if shape:
        return f'pyxhdl.float_array{len(shape)}d{adims}({fspec.exp} downto {-fspec.mant})'
      else:
        return f'float({fspec.exp} downto {-fspec.mant})'

    fatal(f'Unknown type: {dtype}', exc=TypeError)

  def _resize_bits(self, value, nbits):
    if value.dtype.nbits == nbits:
      return value

    xvalue = self.svalue(value)
    result = f'pyxhdl.bits_resize({xvalue}, {nbits})'

    shape = list(value.dtype.shape)
    shape[-1] = nbits

    return value.new_value(result, shape=shape)

  def _to_bool(self, value, dtype):
    if isinstance(value, Value):
      xvalue = self.svalue(value)
      if (isinstance(value.dtype, (Sint, Uint)) or
          (isinstance(value.dtype, Bits) and value.dtype.nbits > 1)):
        return f'and({xvalue})'

      zero = self._cast(0, value.dtype)
      return f'{paren(xvalue)} /= {zero}'

    if isinstance(value, str):
      value = pycu.infer_value(value)

    return 'true' if value else 'false'

  def _literal_bits(self, value, dtype):
    if isinstance(value, str):
      if value.startswith('0b'):
        value = value[2: ]
      if dtype.nbits > len(value):
        bvalue = ('0' * (dtype.nbits - len(value))) + value
      else:
        bvalue = value[-dtype.nbits: ]

      return bvalue.upper()
    elif isinstance(value, int) and abs(value) >= 2**31:
      return self._int2bits(value, dtype.nbits)

  def _to_int(self, value, dtype, itype):
    if isinstance(value, Value):
      if isinstance(value.dtype, (Uint, Sint)):
        xvalue = self.svalue(value)
        rsvalue = f'resize({xvalue}, {dtype.nbits})' if value.dtype.nbits != dtype.nbits else xvalue
        return f'{itype}({rsvalue})' if type(dtype) != type(value.dtype) else rsvalue
      elif isinstance(value.dtype, Bits):
        xvalue = self.svalue(value)
        return f'pyxhdl.cvt_{itype}({xvalue}, {dtype.nbits})'
      elif isinstance(value.dtype, Integer):
        return f'to_{itype}({self.svalue(value)}, {dtype.nbits})'
      elif isinstance(value.dtype, Real):
        return f'to_{itype}(integer({self.svalue(value)}), {dtype.nbits})'
      elif isinstance(value.dtype, Bool):
        zero = f'to_{itype}(0, {dtype.nbits})'
        one = f'to_{itype}(1, {dtype.nbits})'
        xvalue = self.svalue(value)
        return f'pyxhdl.{dtype.name}_ifexp({xvalue}, {one}, {zero})'

    if bvalue := self._literal_bits(value, dtype):
      return (f'pyxhdl.cvt_{itype}("{bvalue}", {dtype.nbits})' if dtype.nbits != 1
              else f'\'{bvalue}\'')

    return (f'to_{itype}({self.svalue(value)}, {dtype.nbits})' if dtype.nbits != 1
            else f'\'{self.svalue(value)}\'')

  def _to_uint(self, value, dtype):
    return self._to_int(value, dtype, 'unsigned')

  def _to_sint(self, value, dtype):
    return self._to_int(value, dtype, 'signed')

  def _try_convert_literal(self, value):
    # This API should not try to convert Python `int` or `float` literals, as those
    # are better off directly going to the cast operators, resulting in less bits
    # resize operations.
    if isinstance(value, bool):
      return Value(BOOL, 'true' if value else 'false')
    if value is None:
      return Value(VOID)
    if isinstance(value, str):
      bvalue = bitstring(value)
      if bvalue is not None:
        return bvalue.new_value(f'"{bvalue.value}"' if bvalue.dtype.nbits > 1
                                else f'\'{bvalue.value}\'')

      dtype, ivalue = self._match_intstring(value)
      if dtype is not None:
        if isinstance(dtype, Uint):
          return Value(dtype, f'to_unsigned({ivalue}, {dtype.nbits})')
        else:
          return Value(dtype, f'to_signed({ivalue}, {dtype.nbits})')

    return value

  def _to_bits(self, value, dtype):
    if isinstance(value, Value):
      if isinstance(value.dtype, (Uint, Sint)):
        xvalue = self.svalue(value)
        if value.dtype.nbits != dtype.nbits:
          xvalue = f'resize({xvalue}, {dtype.nbits})'
        return f'std_logic_vector({xvalue})' if dtype.nbits > 1 else xvalue
      elif isinstance(value.dtype, Bits):
        value = self._resize_bits(value, dtype.nbits)
        return self.svalue(value)
      elif isinstance(value.dtype, Integer):
        return f'std_logic_vector(to_unsigned({self.svalue(value)}, {dtype.nbits}))'
      elif isinstance(value.dtype, Real):
        return f'std_logic_vector(to_unsigned(integer({self.svalue(value)}), {dtype.nbits}))'
      elif isinstance(value.dtype, Bool):
        xvalue = self.svalue(value)
        result = f'pyxhdl.{dtype.name}_ifexp({xvalue}, \'1\', \'0\')'
        if dtype.nbits > 1:
          result = '"' + ('0' * (dtype.nbits - 1)) + '" & ' + result

        return result
      else:
        fatal(f'Unable to convert to {dtype}: {value.dtype}')

    if bvalue := self._literal_bits(value, dtype):
      return f'"{bvalue}"' if dtype.nbits != 1 else f'\'{bvalue}\''

    return (f'std_logic_vector(to_unsigned({self.svalue(value)}, {dtype.nbits}))'
            if dtype.nbits != 1 else f'\'{self.svalue(value)}\'')

  def _to_float(self, value, dtype):
    fspec = self.float_spec(dtype)
    if isinstance(value, Value):
      if isinstance(value.dtype, Float):
        return f'resize({self.svalue(value)}, {fspec.exp}, {fspec.mant})'
      elif isinstance(value.dtype, Bool):
        xvalue = f'pyxhdl.real_ifexp({self.svalue(value)}, 1.0, 0.0)'

    return f'to_float({value}, {fspec.exp}, {fspec.mant})'

  def _to_integer(self, value, dtype):
    if isinstance(value, Value):
      if value.dtype == dtype:
        return self.svalue(value)
      elif isinstance(value.dtype, (Bool, Sint, Uint, Float)):
        return f'to_integer({self.svalue(value)})'
      elif isinstance(value.dtype, Bits):
        return f'to_integer(unsigned({self.svalue(value)}))'
      elif isinstance(value.dtype, Real):
        return f'integer({self.svalue(value)})'
      else:
        fatal(f'Unable to convert to integer: {value} {dtype}')

    return str(value) if isinstance(value, int) else f'integer({value})'

  def _to_real(self, value, dtype):
    if isinstance(value, Value):
      if value.dtype == dtype:
        return self.svalue(value)
      elif isinstance(value.dtype, (Bool, Sint, Uint, Float)):
        return f'to_real({self.svalue(value)})'
      elif isinstance(value.dtype, Bits):
        return f'to_real(unsigned({self.svalue(value)}))'
      elif isinstance(value.dtype, Integer):
        return f'real({self.svalue(value)})'
      else:
        fatal(f'Unable to convert to real: {value} {dtype}')

    return str(value) if isinstance(value, float) else f'real({value})'

  def _scalar_cast(self, value, dtype):
    if isinstance(dtype, Bool):
      return self._to_bool(value, dtype)
    if isinstance(dtype, Uint):
      return self._to_uint(value, dtype)
    if isinstance(dtype, Sint):
      return self._to_sint(value, dtype)
    if isinstance(dtype, Bits):
      return self._to_bits(value, dtype)
    if isinstance(dtype, Float):
      return self._to_float(value, dtype)
    if isinstance(dtype, Integer):
      return self._to_integer(value, dtype)
    if isinstance(dtype, Real):
      return self._to_real(value, dtype)

    fatal(f'Unknown type: {dtype}')

  def _do_cast(self, value, dtype):
    if not isinstance(value, Value):
      value = self._try_convert_literal(value)

    if isinstance(value, np.ndarray):
      shape = dtype.array_shape
      if shape != tuple(value.shape):
        fatal(f'Shape mismatch: {tuple(value.shape)} vs. {shape}')

      element_type = dtype.element_type()
      parts = []
      for idx in np.ndindex(shape):
        parts.append(self._scalar_cast(value[idx].item(), element_type))

      xvalue = flat2shape(parts, shape, '(', ')')
    elif isinstance(value, Value) and value.dtype.array_shape:
      shape, vshape = dtype.array_shape, value.dtype.array_shape
      if shape != vshape:
        fatal(f'Shape mismatch: {vshape} vs. {shape}')

      element_type = dtype.element_type()
      velement_type = value.dtype.element_type()
      avalue = paren(self.svalue(value))
      parts = []
      for idx in np.ndindex(shape):
        substr = ''.join(f'({x})' for x in idx)
        svalue = Value(velement_type, f'{avalue}{substr}')
        parts.append(self._scalar_cast(svalue, element_type))

      xvalue = flat2shape(parts, shape, '(', ')')
    else:
      xvalue = self._scalar_cast(unwrap(value), dtype)
      for _ in range(len(dtype.array_shape)):
        xvalue = f'(others => {xvalue})'

    return xvalue

  def _cast(self, value, dtype):
    if isinstance(value, Value) and value.dtype == dtype:
      return self.svalue(value)

    return self._do_cast(value, dtype)

  def cast(self, value, dtype, isreg=None):
    if isinstance(value, Value) and value.dtype == dtype:
      return value

    xvalue = self._do_cast(value, dtype)
    if isreg is None and isinstance(value, Value):
      isreg = value.isreg

    return Value(dtype, xvalue, isreg=isreg)

  def eval_tostring(self, value):
    xvalue = self.svalue(value)
    if isinstance(value, Value):
      if isinstance(value.dtype, (Uint, Sint)):
        return f'to_hstring({xvalue})'
      if isinstance(value.dtype, (Bool, Bits, Integer, Real)):
        return f'to_string({xvalue})'
      if isinstance(value.dtype, Float):
        return f'to_string(to_real({xvalue}))'

    return self.quote_string(xvalue)

  def quote_string(self, s):
    es = s.replace('"', '""')
    return f'"{es}"'

  def eval_token(self, token):
    if token == 'NOW':
      return f'to_string(now, 1 {self.time_unit()})'

  def emit_finish(self):
    self.emit_code('std.env.finish;')

  def emit_wait_for(self, ts=None):
    if ts is not None:
      wts, tu = self._normalize_time(ts)
      self.emit_code(f'wait for {round(wts)} {tu};')
    else:
      self.emit_code('wait;')

  def emit_wait_rising(self, arg):
    self.emit_code(f'wait until rising_edge({self.svalue(arg)});')

  def emit_wait_falling(self, arg):
    self.emit_code(f'wait until falling_edge({self.svalue(arg)});')

  def emit_wait_until(self, arg):
    self.emit_code(f'wait until {paren(self.svalue(arg))};')

  def emit_report(self, parts, severity=None):
    self._emit_line('report ' + ' & '.join(parts) + ';')

  def emit_write(self, parts):
    self._emit_line('write(output, ' + ' & '.join(parts) + ' & LF);')

  def _gen_array_access(self, arg, idx):
    idx = pyu.as_sequence(idx)

    ashape = arg.dtype.shape
    if len(idx) > len(ashape):
      fatal(f'Wrong indexing for shape: {idx} vs. {ashape}')

    shape, coords = [], []
    for i, ix in enumerate(idx):
      if isinstance(ix, slice):
        if isinstance(ix.start, Value):
          if ix.stop is not None:
            fatal(f'Variable part select ({arg} [{i}]) slice stop must be empty: {ix.stop}')
          if ix.step is None:
            fatal(f'Variable part select ({arg} [{i}]) slice step must not be empty')

          step = abs(ix.step)
          if step > ashape[i]:
            fatal(f'Variable part select ({arg} [{i}]) is too big: {step} ({ashape[i]})')

          base = self._to_integer(ix.start, Integer())

          if (i == len(ashape) - 1) == (ix.step >= 0):
            coords.append(f'({paren(base)} + {step - 1}) downto {paren(base)}')
          else:
            coords.append(f'{paren(base)} to ({paren(base)} + {step - 1})')

          shape.append(step)
        else:
          step = ix.step if ix.step is not None else 1
          if abs(step) != 1:
            fatal(f'Slice step must be 1: {step}')

          start, stop = validate_slice(ix.start, ix.stop, ashape[i], step)

          slice_dir = {1: 'downto', -1: 'to'}
          if i == len(ashape) - 1:
            coords.append(f'{stop - step} {slice_dir[step]} {start}')
          else:
            coords.append(f'{start} {slice_dir[step]} {stop - step}')

          shape.append(abs(stop - start))
      else:
        if isinstance(ix, Value):
          if not isinstance(ix.dtype, Integer):
            ix = self._to_integer(ix, Integer())
        elif isinstance(ix, int) and ix < 0:
          ix = ashape[i] + ix

        coords.append(self.svalue(ix))
        shape.append(1)

    shape = pyu.squeeze(shape + list(arg.dtype.full_shape[len(idx): ]), keep_dims=1,
                        sdir=pyu.MAJOR)
    avalue = self.svalue(arg) + ''.join(f'({x})' for x in coords)

    return avalue, shape

  def flush(self):
    return self._load_libs() + self._expand()

  def is_root_variable(self, var):
    return var.isreg or var.is_const()

  def var_remap(self, var, is_store):
    return var

  def _emit_attributes(self, name, varkind, attributes):
    def declare_attr(aname, atype):
      rtype = self._mod_attributes.get(aname)
      if rtype is None:
        self._mod_attributes[aname] = atype
        self._emit_line(f'attribute {aname} : {atype};')
      elif atype != rtype:
        fatal(f'Attribute {aname} was {rtype} before, now {atype}')


    for aname, avalue in self._enum_attributes(attributes):
      if isinstance(avalue, str):
        declare_attr(aname, 'string')
        self._emit_line(f'attribute {aname} of {name} : {varkind} is "{avalue}";')
      elif isinstance(avalue, int):
        declare_attr(aname, 'integer')
        self._emit_line(f'attribute {aname} of {name} : {varkind} is {avalue};')
      elif isinstance(avalue, float):
        declare_attr(aname, 'real')
        self._emit_line(f'attribute {aname} of {name} : {varkind} is {avalue};')
      else:
        fatal(f'Unknown attribute type: {value}')

  def emit_declare_variable(self, name, var):
    vtype = self._type_of(var.dtype)

    if var.is_const():
      vprefix = 'constant'
    elif self._proc.kind == ROOT_PROCESS:
      vprefix = 'signal'
    else:
      vprefix = 'variable' if var.isreg is False else 'signal'

    if var.init is not None:
      xinit = self._cast(var.init, var.dtype)
      vinit = f' := {xinit}'
    else:
      vinit = ''

    self._emit_line(f'{vprefix} {name} : {vtype}{vinit};')

    vspec = var.vspec
    if vspec is not None and vspec.attributes is not None:
      self._emit_attributes(name, vprefix, vspec.attributes)

  def emit_assign(self, var, name, value):
    xvalue = self._cast(value, var.dtype)

    xdelay, xtrans = '', ''

    delay = self.get_context('delay')
    trans = self.get_context('trans')
    if delay is not None or trans is True:
      # Delays apply only to signals, not registers ...
      if var.isreg is False:
        fatal(f'Cannot use delay/trans on wires: {var}')
      if self._proc.kind == ROOT_PROCESS:
        fatal(f'Cannot use delay/trans within a root process: {var}')

      if delay is not None:
        xdelay = f' after {self.svalue(delay)} {self.time_unit()}'
      if trans is True:
        xtrans = 'transport '

    vspec = var.vspec
    if (vspec is not None and vspec.port is not None and vspec.port.is_wr()):
      asop = '<='
    elif self._proc.kind == ROOT_PROCESS:
      asop = '<='
    else:
      asop = ':=' if var.isreg is False else '<='

    self._emit_line(f'{var.value} {asop} {xtrans}{xvalue}{xdelay};')

  def make_port_arg(self, port_arg):
    return port_arg.new_isreg(False)

  def emit_entity(self, ent, kwargs, ent_name=None):
    if ent_name is None:
      ent_name = pyiu.cname(ent)

    iname = self._get_entity_inst(ent_name)

    with self.placement(self._entity_place):
      self._emit_line(f'{iname} : entity {ent_name}')

      eparams = kwargs.pop(PARAM_KEY, None)
      if eparams is not None:
        self._emit_line(f'generic map (')
        with self.indent():
          self._emit_lines((f'{k} => {v}' for k, v in eparams.items()), sep=',')
        self._emit_line(f')')

      self._emit_line(f'port map (')
      with self.indent():
        binds = []
        for i, pin in enumerate(ent.PORTS):
          arg = kwargs[pin.name]

          if pin.is_ifc():
            xargs = pin.ifc_expand(arg)
          else:
            xargs = ((pin, arg),)

          for xpin, xarg in xargs:
            if not isinstance(xarg, Value):
              fatal(f'Argument must be a Value subclass: {xarg}')

            if xarg.is_none():
              binds.append(f'{xpin.name} => open')
            else:
              earg = self.svalue(xarg)
              binds.append(f'{xpin.name} => {earg}')

        self._emit_lines(binds, sep=',')

      self._emit_line(f');')

  def emit_module_def(self, name, ent, comment=None):
    self._emit_header()

    self._mod_comment = comment
    if comment:
      self.emit_comment(comment)
    self._emit_line(f'entity {name} is')
    if ent.args:
      with self.indent():
        self._emit_line(f'port (')
        with self.indent():
          xports = []
          for ap in ent.expanded_ports():
            pin, arg = ap.port, ap.arg

            pdir = 'in' if pin.is_ro() else 'out' if pin.is_wo() else 'inout'
            ptype = self._type_of(arg.dtype)
            xports.append(f'{pin.name} : {pdir} {ptype}')

          self._emit_lines(xports, sep=';')

        self._emit_line(f');')

    self._emit_line(f'end entity;')

  def emit_module_decl(self, name, ent):
    self._emit_header()

    if self._mod_comment:
      self.emit_comment(self._mod_comment)
    self._emit_line(f'architecture {self._arch} of {name} is')
    self.module_vars_place = self.emit_placement(extra_indent=1)

    self._emit_line(f'begin')

    self._modules_place = self.emit_placement(extra_indent=1)
    self._entity_place = self.emit_placement(extra_indent=1)

  def emit_module_end(self):
    with self.placement(self._modules_place):
      for iid, inst in self._itor:
        self._emit_line(f'{iid} : entity {inst.name}')
        if inst.params:
          self._emit_line('generic map (')
          with self.indent():
            self._emit_lines((f'{k} => {v}' for k, v in inst.params.items()), sep=',')
          self._emit_line(')')

        self._emit_line(f'port map (')
        with self.indent():
          self._emit_lines((f'{k} => {self.svalue(v)}' for k, v in inst.args.items()), sep=',')
        self._emit_line(');')

    self._emit_line(f'end architecture;')

    self._mod_attributes = dict()
    self._module_reset()

  def emit_process_decl(self, name, sensitivity=None, process_kind=None,
                        process_args=None):
    self._process_init(name, process_kind, process_args, sensitivity)

    proc_decl = f'{name} : process'
    if sensitivity:
      proc_decl += ' (' + ', '.join(name for name in sensitivity.keys()) + ')'
    else:
      proc_mode = process_args.get('proc_mode') if process_args else None
      if proc_mode == 'comb':
        proc_decl += ' (all)'

    self._emit_line(proc_decl)
    self.process_vars_place = self.emit_placement(extra_indent=1)

  def emit_process_begin(self):
    self._emit_line(f'begin')
    if self._proc.sens:
      tests = []
      for name, sens in self._proc.sens.items():
        if sens.trigger == POSEDGE:
          tests.append(f'rising_edge({name})')
        elif sens.trigger == NEGEDGE:
          tests.append(f'falling_edge({name})')

      if tests:
        self._proc_indent += 1
        self._indent += 1
        self._emit_line('if ' + ' or '.join(tests) + ' then')

  def emit_process_end(self):
    # There is no special "init" process in VHDL, so we need to make sure that once
    # it ran, it stops.
    if self._proc.kind == INIT_PROCESS:
      with self.indent():
        self._emit_line('wait;')

    if self._proc_indent > 0:
      self._proc_indent -= 1
      self._emit_line('end if;')
      self._indent -= 1

    self._emit_line(f'end process;')

    self._process_reset()

  def emit_comment(self, msg):
    for ln in msg.split('\n'):
      self._emit_line(f'-- {ln}')

  def emit_If(self, test):
    xtest = self._cast(test, BOOL)
    self._emit_line(f'if {xtest} then')

  def emit_Elif(self, test):
    xtest = self._cast(test, BOOL)
    self._emit_line(f'elsif {xtest} then')

  def emit_Else(self):
    self._emit_line(f'else')

  def emit_EndIf(self):
    self._emit_line(f'end if;')

  def emit_For(self, vname, start, end, step):
    loop_dir = 'to' if step > 0 else 'downto'

    if self._proc.kind == ROOT_PROCESS:
      gname = self._revgen.newname('genfor')
      self._emit_line(f'{gname}: for {vname} in {start} {loop_dir} {end} generate')
    else:
      self._emit_line(f'for {vname} in {start} {loop_dir} {end} loop')

  def emit_EndFor(self):
    if self._proc.kind == ROOT_PROCESS:
      self._emit_line(f'end generate;')
    else:
      self._emit_line(f'end loop;')

  def emit_Break(self):
    self._emit_line(f'exit;')

  def emit_Continue(self):
    self._emit_line(f'next;')

  def emit_Assert(self, test, parts):
    xtest = self.svalue(test)
    if parts:
      self._emit_line(f'assert {xtest} report ' + ' & '.join(parts) + ';')
    else:
      self._emit_line(f'assert {xtest};')

  def _null_pad(self, indent=None, size=None, **kwargs):
    spaces = ' ' * (self._indent_spaces * indent)

    return () if size > 1 else (f'{spaces}null;',)

  def emit_match_cases(self, subject, cases):
    xsubject = self.svalue(subject)
    self._emit_line(f'case {paren(xsubject)} is')

    with self.indent():
      for mc in cases:
        if mc.pattern is not None:
          xpattern = self._cast(mc.pattern, subject.dtype)
          self._emit_line(f'when {paren(xpattern)} =>')
        else:
          self._emit_line(f'when others =>')

        if len(mc.scope) == 0:
          mc.scope.append(functools.partial(self._null_pad, indent=mc.scope.indent))

        self._emit(mc.scope)

    self._emit_line(f'end case;')

  def _build_op(self, op, *args):
    sop = _OPSYMS[pyiu.classof(op)]
    if sop.isfn:
      arglist = ', '.join(paren(arg) for arg in args)

      return f'{sop.sym}({arglist})'
    else:
      opsep = f' {sop.sym} '

      return opsep.join(paren(arg) for arg in args)

  def _arith_ctype_cast(self, arg, dtype):
    if isinstance(arg, int):
      if abs(arg) >= 2**31 or type(dtype) not in {Uint, Sint, Float, Real, Integer}:
        return Value(dtype, self._cast(arg, dtype))
    elif isinstance(arg, float):
      if type(dtype) not in {Float, Real}:
        return Value(dtype, self._cast(arg, dtype))

    return Value(dtype, dtype.ctype(arg))

  def eval_BinOp(self, op, left, right):
    if isinstance(op, (ast.Add, ast.Sub, ast.Mult, ast.Div, ast.Mod)):
      left, right = self._marshal_arith_op([left, right],
                                           ctype_cast=self._arith_ctype_cast)
      xleft, xright = self.svalue(left), self.svalue(right)

      alog.debug(lambda: f'\tBinOp: {xleft}\t{pyiu.cname(op)}\t{xright}')
      result = self._build_op(op, xleft, xright)
      # The signed/unsigned multiplication result has a number of bits which is the
      # sum of the ones of the operands, which is not the behaviour we want.
      if isinstance(op, ast.Mult) and isinstance(left.dtype, (Sint, Uint)):
        result = f'resize({result}, {left.dtype.nbits})'

      return Value(left.dtype, result)
    elif isinstance(op, (ast.LShift, ast.RShift)):
      left, right = self._marshal_shift_op(left, right)
      xleft, xright = self.svalue(left), self.svalue(right)

      alog.debug(lambda: f'\tBinOp: {xleft}\t{pyiu.cname(op)}\t{xright}')

      return Value(left.dtype, self._build_op(op, xleft, xright))
    elif isinstance(op, (ast.BitOr, ast.BitXor, ast.BitAnd)):
      left, right = self._marshal_bit_op([left, right])
      xleft, xright = self.svalue(left), self.svalue(right)

      alog.debug(lambda: f'\tBinOp: {xleft}\t{pyiu.cname(op)}\t{xright}')

      return Value(left.dtype, self._build_op(op, xleft, xright))
    elif isinstance(op, ast.MatMult):
      # Steal MatMult ('@') for concatenation!
      dtype, (left, right) = self._marshal_concat_op([left, right])
      xleft, xright = self.svalue(left), self.svalue(right)

      alog.debug(lambda: f'\tBinOp: {xleft}\t{pyiu.cname(op)}\t{xright}')

      return Value(dtype, self._build_op(op, xleft, xright))
    else:
      fatal(f'Unsupported operation: {op}')

  def eval_UnaryOp(self, op, arg):
    xvalue = self.svalue(arg)

    alog.debug(lambda: f'\tUnaryOp: {pyiu.cname(op)}\t{xvalue}')

    if isinstance(op, ast.UAdd):
      # Unary addition is a noop for HDL data types.
      result = xvalue
    elif isinstance(op, ast.USub):
      result = f'-{paren(xvalue)}'
    elif isinstance(op, (ast.Not, ast.Invert)):
      result = f'not {paren(xvalue)}'
    else:
      fatal(f'Unsupported operation: {op}')

    return Value(arg.dtype, result)

  def eval_BoolOp(self, op, args):
    xargs = [self._cast(a, BOOL) for a in args]

    alog.debug(lambda: f'\tBoolOp: {pyiu.cname(op)}\t{pyu.stri(xargs)}')

    if isinstance(op, ast.And):
      result = self._paren_join(' and ', xargs)
    elif isinstance(op, ast.Or):
      result = self._paren_join(' or ', xargs)
    else:
      fatal(f'Unsupported operation: {op}')

    return Value(BOOL, result)

  def eval_Compare(self, left, ops, comps):
    comps = self._marshal_compare_op([left] + list(comps))
    xcomps = [self.svalue(comp) for comp in comps]

    alog.debug(lambda: f'\tCompare: {[pyiu.cname(x) for x in ops]}\t{pyu.stri(xcomps)}')

    results = []
    for i, op in enumerate(ops):
      cres = self._build_op(op, xcomps[i], xcomps[i + 1])
      results.append(cres)

    result = self._paren_join(' and ', results)

    return Value(BOOL, result)

  def eval_Subscript(self, arg, idx):
    result, shape = self._gen_array_access(arg, idx)

    return arg.new_value(result, shape=shape, keepref=True)

  def eval_IfExp(self, test, body, orelse):
    xtest = self.svalue(test)
    body, orelse = self._marshal_ifexp_op([body, orelse])
    xbody, xorelse = self.svalue(body), self.svalue(orelse)

    alog.debug(lambda: f'\tIfExp: {xtest} ? {xbody} : {xorelse}')

    result = f'pyxhdl.{body.dtype.name}_ifexp({xtest}, {xbody}, {xorelse})'

    return Value(body.dtype, result)

  # Extension functions.
  def eval_is_nan(self, value):
    if not isinstance(value.dtype, Float):
      fatal(f'Unsupported type: {value.dtype}')

    result = f'Isnan({self.svalue(value)})'

    return Value(BOOL, result)

  def eval_is_inf(self, value):
    if not isinstance(value.dtype, Float):
      fatal(f'Unsupported type: {value.dtype}')

    result = f'not Finite({self.svalue(value)})'

    return Value(BOOL, result)


# Register VHDL (VHDL >= 2008) emitter class.
Emitter.register('vhdl', VHDL_Emitter)




#!/usr/bin/env python3

from setuptools import setup, find_packages


setup(name='pyxhdl',
      version='0.35',
      description='Use Python as HDL language, and generate equivalent SystemVerilog and VHDL code',
      author='Davide Libenzi',
      packages=find_packages(),
      package_data={
          'pyxhdl': ['hdl_libs/**', ],
      },
      python_requires='>=3.9',
      install_requires=[
          'numpy',
          'python_misc_utils',
      ],
      )



import re

import py_misc_utils.core_utils as pycu
import py_misc_utils.inspect_utils as pyiu
import py_misc_utils.utils as pyu

from .decorators import *
from .port import *
from .types import *
from .vars import *
from .utils import *


class _InterfaceBase:

  def __init__(self, name, **kwargs):
    from . import xlib

    self.name = name
    self.args = tuple(sorted(kwargs.keys()))
    self.fields = dict()
    self.xlib = xlib

    for k, v in kwargs.items():
      setattr(self, k, v)

  @property
  def origin(self):
    return getattr(self, 'origin_ifc', self)

  def __hash__(self):
    hargs = tuple((getattr(self, name) for name in self.args))
    hfields = tuple((name, getattr(self, name).dtype) for name in
                    sorted(self.fields.keys()))

    return hash(hargs + hfields)

  def __eq__(self, other):
    if self.args != other.args:
      return False

    if self.fields != other.fields:
      return False

    for name in self.args:
      if getattr(self, name) != getattr(other, name):
        return False

    for name in self.fields.keys():
      if getattr(self, name).dtype != getattr(other, name).dtype:
        return False

    return True

  def __repr__(self):
    parts = [f'{name}:{getattr(self, name)}' for name in self.args]
    parts += [f'{name}:{getattr(self, name).dtype}' for name in self.fields.keys()]

    return pyiu.cname(self) + '(' + ', '.join(parts) + ')'


class InterfaceView(_InterfaceBase):

  def __init__(self, origin, name):
    super().__init__(name, **{k: getattr(origin, k) for k in origin.args})
    self.origin_ifc = origin

  def add_field(self, name, value, mode):
    if not isinstance(value, Value):
      fatal(f'Wrong field value type (should be Value): {value}')

    vref = value.ref
    if vref is None:
      fatal(f'Wrong field value type (should contain a Ref): {value}')

    xname = subname(self.name, name)

    vref = vref.new_mode(vref.mode if vref.mode == Ref.RO else mode)
    vref = vref.new_name(xname, vname=xname)

    xvalue = value.new_value(vref)

    setattr(self, name, xvalue)

    self.fields[name] = xname


class Interface(_InterfaceBase):

  def __init__(self, name, **kwargs):
    super().__init__(name, **kwargs)
    self._uname = self.xlib.generate_name(name, shortzero=True)
    if fstr := getattr(self, 'FIELDS', None):
      self.create_fields(fstr)

  def _mkvalue(self, name, value, init=None):
    if isinstance(value, Type):
      if init is not None:
        return mkvreg(value, init, name=name)
      else:
        return mkreg(value, name=name)
    elif isinstance(value, str):
      if init is not None:
        return mkvreg(dtype_from_string(value), init, name=name)
      else:
        return mkreg(dtype_from_string(value), name=name)
    else:
      fatal(f'Invalid interface value: {value}')

  def mkfield(self, name, value, init=None):
    if isinstance(value, Value):
      if value.name is not None:
        xname, fvalue = value.name, value
      else:
        xname, fvalue = subname(self._uname, name), value

        self.xlib.assign(xname, fvalue)
    else:
      xname = subname(self._uname, name)
      fvalue = self._mkvalue(xname, value, init=init)

      self.xlib.assign(xname, fvalue)

    setattr(self, name, self.xlib.load(xname))

    self.fields[name] = xname

  def create_fields(self, fstr):
    for fs in pyu.comma_split(fstr):
      name, fvtype = re.split(r'\s*:\s*', fs, maxsplit=1)
      ftype, *fvalue = re.split(r'\s*=\s*', fvtype, maxsplit=1)

      self.mkfield(name, ftype, init=pycu.infer_value(fvalue[0]) if fvalue else None)

  @hdl
  def reset(self):
    pass

  def create_port_view(self, name, port_name):
    ports_spec = getattr(self, port_name, None)
    if ports_spec is None:
      fatal(f'Invalid port name: {port_name}')

    ports = Port.parse_list(ports_spec)

    view = InterfaceView(self, name)
    for pin in ports:
      mode = Ref.RO if pin.idir == Port.IN else Ref.RW

      view.add_field(pin.name, getattr(self, pin.name), mode)

    return view

  def expand_port(self, name, port_name):
    ports_spec = getattr(self, port_name, None)
    if ports_spec is None:
      fatal(f'Invalid port name: {port_name}')

    ports = Port.parse_list(ports_spec)
    expanded = []
    for pin in ports:
      expanded.append((pycu.new_with(pin, name=subname(name, pin.name)),
                       getattr(self, pin.name)))

    return tuple(expanded)




import argparse
import os

import py_misc_utils.alog as alog
import py_misc_utils.app_main as app_main
import py_misc_utils.gfs as gfs
import py_misc_utils.module_utils as pymu
import py_misc_utils.utils as pyu

from .emitter import *
from .main_utils import *
from .pyxhdl import *
from .types import *
from .utils import *
from .vars import *
from .vhdl_emitter import *

from . import testbench as TB


def _main(args):
  if args.cfgfile is not None:
    parse_args(args.cfgfile, args)

  mod = pymu.load_module(args.input_file)
  ent_class = getattr(mod, args.entity)

  gglobals = create_globals(mod, source_globals=globals())

  ekwargs = parse_kwargs(args.ekwargs, gglobals)
  emitter = Emitter.create(args.backend.lower(),
                           cfg_file=args.emitter_cfgfile,
                           **ekwargs)

  codegen = CodeGen(emitter, gglobals)

  inputs = parse_inputs(args.inputs, args.kwargs, gglobals)

  for pin in ent_class.PORTS:
    if pin.name not in inputs and pin.default is not None:
      if pin.idir == Port.IN:
        inputs[pin.name] = mkwire(pin.default)
      else:
        inputs[pin.name] = mkreg(pin.default)

  with codegen.context():
    if args.testbench:
      TB.generate(codegen, args, ent_class, inputs)
    else:
      codegen.generate_entity(ent_class, inputs)

    code = codegen.flush()

    with gfs.std_open(args.output_file, mode='w') as ofd:
      for ln in code:
        print(ln, file=ofd)


if __name__ == '__main__':
  parser = argparse.ArgumentParser(description='PyXHDL Code Generator',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument('--input_file', required=True,
                      help='The path to the Python source file containing the root entity')
  parser.add_argument('--entity', required=True,
                      help='The root entity name')
  parser.add_argument('--backend', default='verilog',
                      choices=set(Emitter.available()),
                      help='The backend to generate the code for')
  parser.add_argument('--inputs', action='append',
                      help='The inputs for the root entity')
  parser.add_argument('--kwargs', action='append',
                      help='The keyword arguments for the root entity')
  parser.add_argument('--emitter_cfgfile',
                      help='The path to the YAML file containing the emitter configuration')
  parser.add_argument('--ekwargs', action='append',
                      help='The keyword arguments for the emitter')
  parser.add_argument('--cfgfile', default=os.getenv('PYXHDL_CONFIG'),
                      help='The path to the YAML file containing the generator configuration')
  parser.add_argument('--output_file',
                      help='The path to the output file for the generated code (default STDOUT)')
  parser.add_argument('--testbench', action='store_true',
                      help='Run the entity with a testbench')

  TB.add_arguments(parser)

  app_main.main(parser, _main)




import py_misc_utils.inspect_utils as pyiu
import py_misc_utils.utils as pyu

from .decorators import *
from .emitter import *
from .pyxhdl import *
from .types import *
from .vars import *
from .utils import *
from .verilog_emitter import *
from .xcall import *


def comment(comm):
  ctx = CodeGen.current()

  return ctx.emitter.emit_comment(comm)


def register_module(mid, code, replace=None, global_reg=None):
  ctx = CodeGen.current()

  if global_reg is True or ctx is None:
    Emitter.glob_register_module(mid, code, replace=replace)
  else:
    ctx.emitter.register_module(mid, code, replace=replace)


def cast(value, dtype) -> Value:
  ctx = CodeGen.current()

  return ctx.emitter.cast(value, dtype)


def load(name) -> Value:
  ctx = CodeGen.current()

  return ctx.load_var(name)


def assign(name, value):
  ctx = CodeGen.current()

  ctx.assign_value(name, value)


def context(**kwargs):
  ctx = CodeGen.current()

  return ctx.emitter_context(**kwargs)


def no_hdl():
  ctx = CodeGen.current()

  return ctx.no_hdl()


def force_hdl():
  ctx = CodeGen.current()

  return ctx.force_hdl()


def loop_mode_hdl():
  ctx = CodeGen.current()

  return ctx.set_loop_mode('hdl')


def loop_mode_unrolled():
  ctx = CodeGen.current()

  return ctx.set_loop_mode('unrolled')


def finish():
  ctx = CodeGen.current()
  ctx.emitter.emit_finish()


def wait_for(ts=None):
  ctx = CodeGen.current()
  ctx.emitter.emit_wait_for(ts=ts)


def wait_rising(arg):
  ctx = CodeGen.current()
  ctx.emitter.emit_wait_rising(arg)


def wait_falling(arg):
  ctx = CodeGen.current()
  ctx.emitter.emit_wait_falling(arg)


@hdl
def wait_until(arg):
  if not arg:
    with no_hdl():
      ctx = CodeGen.current()
      ctx.emitter.emit_wait_until(arg)


def report(fmt, **kwargs):
  ctx = CodeGen.current()
  ctx.emit_report(fmt, **kwargs)


def write(fmt, **kwargs):
  ctx = CodeGen.current()
  ctx.emit_write(fmt, **kwargs)


def xeval(code, **args):
  ctx = CodeGen.current()

  filename, lineno = pyiu.parent_coords()
  return ctx.run_code(code, args, 'eval', filename=filename, lineno=lineno)


def xexec(code, **args):
  ctx = CodeGen.current()

  filename, lineno = pyiu.parent_coords()
  ctx.run_code(code, args, 'exec', filename=filename, lineno=lineno)


## External Functions
float_equal = create_function(
  'float_equal',
  {
    'vhdl': 'pyxhdl.float_equal',
    'verilog': Verilog_Emitter.fpmod_resolve('vfpu', 'rcloseto', 0),
  },
  fnsig='f*, real, real',
  dtype=BOOL)


real_equal = create_function(
  'real_equal',
  {
    'vhdl': 'pyxhdl.float_equal',
    'verilog': 'pyxhdl::float_equal',
  },
  fnsig='real, real, real',
  dtype=BOOL)


def generate_name(name, shortzero=False):
  ctx = CodeGen.current()

  return ctx.generate_name(name, shortzero=shortzero)


def backend():
  ctx = CodeGen.current()

  return ctx.emitter.kind


def load_extern_module(path):
  ctx = CodeGen.current()

  return ctx.emitter.load_extern_module(path)




import inspect
import re
import traceback

import numpy as np

import py_misc_utils.core_utils as pycu
import py_misc_utils.inspect_utils as pyiu
import py_misc_utils.utils as pyu

from .utils import *


class Type:

  __slots__ = ('name', 'full_shape', 'ctype')

  def __init__(self, name, shape, ctype):
    self.name = name
    self.full_shape = tuple(shape[:])
    self.ctype = ctype

  @property
  def has_bits(self):
    return self.full_shape[-1] is not None

  @property
  def shape(self):
    return self.full_shape if self.has_bits else self.full_shape[: -1]

  @property
  def array_shape(self):
    return self.full_shape[: -1]

  @property
  def ndim(self):
    return len(self.full_shape) - (0 if self.has_bits else 1)

  @property
  def nbits(self):
    return self.full_shape[-1] if self.has_bits else None

  @property
  def size(self):
    return np.prod(self.full_shape[: -1])

  def __hash__(self):
    return hash((self.name, self.full_shape, self.ctype))

  def __eq__(self, other):
    return (self.name == other.name and self.full_shape == other.full_shape and
            self.ctype == other.ctype)

  def __str__(self):
    return f'{self.name}(' + ', '.join(str(x) for x in self.shape) + ')'

  def __repr__(self):
    return f'{pyiu.cname(self)}({self.name}, {self.full_shape}, {self.ctype})'

  def new_shape(self, *shape):
    if not self.has_bits and (not shape or shape[-1] is not None):
      shape = shape + (None,)

    return pycu.new_with(self, full_shape=shape)

  def element_type(self):
    return self.new_shape(*self.full_shape[-1: ])


class Sint(Type):

  def __init__(self, *shape):
    super().__init__('sint', shape, int)


class Uint(Type):

  def __init__(self, *shape):
    super().__init__('uint', shape, int)


class Bits(Type):

  def __init__(self, *shape):
    super().__init__('bits', shape, int)


class Float(Type):

  def __init__(self, *shape):
    super().__init__('float', shape, float)


class Bool(Type):

  def __init__(self, *shape):
    super().__init__('bool', shape + (None,), bool)


class Integer(Type):

  def __init__(self, *shape):
    super().__init__('integer', shape + (None,), int)


class Real(Type):

  def __init__(self, *shape):
    super().__init__('real', shape + (None,), float)


class Void(Type):

  def __init__(self):
    super().__init__('void', (None,), None.__class__)


INT4 = Sint(4)
INT8 = Sint(8)
INT16 = Sint(16)
INT32 = Sint(32)
INT64 = Sint(64)
INT128 = Sint(128)

UINT4 = Uint(4)
UINT8 = Uint(8)
UINT16 = Uint(16)
UINT32 = Uint(32)
UINT64 = Uint(64)
UINT128 = Uint(128)

FLOAT16 = Float(16)
FLOAT32 = Float(32)
FLOAT64 = Float(64)
FLOAT80 = Float(80)
FLOAT128 = Float(128)

BOOL = Bool()
BIT = Bits(1)
INT = Integer()
REAL = Real()
VOID = Void()


_TYPE_CLASS = dict(s=Sint,
                   u=Uint,
                   b=Bits,
                   f=Float)
_TYPE_NMAP = dict(bit=BIT,
                  int=INT,
                  real=REAL,
                  bool=BOOL,
                  void=VOID)
_HAS_BITS = {Uint, Sint, Bits, Float}


def has_bits(tclass):
  return tclass in _HAS_BITS


def tclass_from_string(tcls):
  tclass = _TYPE_CLASS.get(tcls)
  if tclass is None:
    fatal(f'Unknown type class: {tcls}')

  return tclass


def dtype_from_string(s):
  # Accepted formats:
  #
  #   u32
  #   real
  #   u16(4, 4)
  #   f32(12)
  ls = s.lower()

  m = re.match(r'(' + '|'.join(_TYPE_CLASS.keys()) + r')(\d+)', ls)
  if m:
    dtype = _TYPE_CLASS[m.group(1)](int(m.group(2)))
    ls = ls[m.end(): ]
  else:
    m = re.match(r'(' + '|'.join(_TYPE_NMAP.keys()) + r')', ls)
    if m:
      dtype = _TYPE_NMAP[m.group(1)]
      ls = ls[m.end(): ]
    else:
      fatal(f'Unknown type string: {s}')

  m = re.match(r'\((\d+(,\d+)*)\)$', ls)
  if m:
    shape = tuple(int(x) for x in m.group(1).split(','))
  elif not ls:
    shape = None
  else:
    fatal(f'Unknown type string: {s}')

  return mkarray(dtype, *shape) if shape else dtype


class TypeMatcher:

  def __init__(self, dtype=None, tclass=None):
    self.dtype = dtype or ()
    self.tclass = tclass or ()

  @classmethod
  def parse(cls, tmstr):
    dtype, tclass = [], []
    for tstr in pyu.resplit(tmstr, ','):
      if tstr != '*':
        m = re.match(r'(.+)\*$', tstr)
        if m:
          tclass.append(tclass_from_string(m.group(1)))
        else:
          dtype.append(dtype_from_string(tstr))

    return cls(dtype=tuple(dtype), tclass=tuple(tclass))

  def check_value(self, arg, msg=''):
    if self.dtype:
      matched = False
      for dtype in self.dtype:
        if dtype == arg.dtype:
          matched = True
          break

      if not matched:
        fatal(f'Mismatch type{msg}: {arg.dtype} vs. {self.dtype}')

    if self.tclass:
      matched = False
      for tclass in self.tclass:
        if isinstance(arg.dtype, tclass):
          matched = True
          break

      if not matched:
        fatal(f'Mismatch type class{msg}: {arg.dtype} vs. ' \
              f'{tuple(pyiu.cname(x) for x in self.tclass)}')

  def cast(self, arg, dtype_fn=None, tclass_fn=None):
    cex = []
    if dtype_fn is not None and self.dtype:
      for dtype in self.dtype:
        try:
          return dtype_fn(arg, dtype)
        except Exception as ex:
          cex.append(ex)

    if tclass_fn is not None and self.tclass:
      for tclass in self.tclass:
        try:
          return tclass_fn(arg, tclass)
        except Exception as ex:
          cex.append(ex)

    if cex:
      xmsgs = []
      map(lambda x: xmsgs.extend(traceback.format_exception(x)), cex)

      fatal('\n'.join(xmsgs), exc=ValueError)

    return arg


def mkarray(base_type, *shape):
  ashape = shape + base_type.full_shape

  return base_type.new_shape(*ashape)

